@@ -196,27 +196,6 @@ printline(FILE *f, unsigned char *line, int w)
 	fprintf(f, "\n");
 }
 
-static inline int
-getrun(const unsigned char *line, int x, int w, int c)
-{
-	int z;
-	int b;
-
-	if (x < 0)
-		x = 0;
-
-	z = x;
-	while (z < w)
-	{
-		b = getbit(line, z);
-		if (c != b)
-			break;
-		z ++;
-	}
-
-	return z - x;
-}
-
 static inline int
 findchanging(const unsigned char *line, int x, int w)
 {
@@ -297,14 +276,14 @@ setbits(unsigned char *line, int x0, int x1)
 
 typedef struct fz_faxd_s fz_faxd;
 
-typedef enum fax_stage_e
+enum
 {
 	SNORMAL,	/* neutral state, waiting for any code */
 	SMAKEUP,	/* got a 1d makeup code, waiting for terminating code */
 	SEOL,		/* at eol, needs output buffer space */
 	SH1, SH2,	/* in H part 1 and 2 (both makeup and terminating codes) */
 	SDONE		/* all done */
-} fax_stage_e;
+};
 
 struct fz_faxd_s
 {
@@ -324,7 +303,7 @@ struct fz_faxd_s
 	int bidx;
 	unsigned int word;
 
-	fax_stage_e stage;
+	int stage;
 
 	int a, c, dim, eolc;
 	unsigned char *ref;
